# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import functools
import glob
import pathlib
import shutil
import subprocess  # nosec
import urllib
from contextlib import closing
from pathlib import Path
from shutil import copyfile
from typing import Union

from soufi import exceptions, finder

API_TIMEOUT = 30  # seconds
# Shell snippet that will source an APKBUILD and spit out the vars that we
# need. CARCH is set to work around a bug in 3.12.0 scripts. CTARGET_ARCH
# is needed for more complex scripts like `community/go`.
SH_GET_APK_VARS = """
function die () {{
:
}}
export CARCH=$(arch)
export CTARGET_ARCH=$(arch)
. {file}
echo $source
echo $subpackages
echo $provides
echo $pkgver
echo $pkgrel
"""


class AlpineFinder(finder.SourceFinder):
    """Find Alpine APK sources.

    :param aports_dir: An aports branch checked out locally.

    It is recommended to clone the branch like this to save time/space:
    git clone --depth 1 --branch vN.N.N git://git.alpinelinux.org/aports

    The results of parsing each directory with the branch is cached, so it
    is also recommended to check out each branch into its own place.

    The caller therefore must know in which release branch the package is
    likely to reside as it is impossible to know this information outside
    the context of an actual distro image.  This Finder *could* iterate
    over every release, but this is infefficient. There's nothing
    stopping the caller doing this though, as required.
    """

    distro = finder.Distro.alpine.value

    # See https://wiki.alpinelinux.org/wiki/APKBUILD_Reference
    # for info on dealing with APKBUILD files.

    def __init__(self, *args, aports_dir, **kwargs):
        # TODO: take a sudo user name that will be used to run APKBUILDs
        self.aports_dir = aports_dir
        super().__init__(*args, **kwargs)

    @functools.lru_cache(maxsize=128)
    def _find_apkbuilds(self, aports_dir):
        """Given a starting dir, find all the APKBUILD files.

        :return: A dict mapping package names to the APKBUILD path.
        """
        apkbuilds = {}
        for apkbuild in glob.iglob(
            f"{aports_dir}/**/*/APKBUILD", recursive=True
        ):
            path = Path(apkbuild)
            package = path.parent.name
            apkbuilds[package] = path
        return apkbuilds

    def _find_apkbuild_path(self, name):
        """Return the Path to the APKBUILD for package with 'name'."""
        apkbuilds = self._find_apkbuilds(self.aports_dir)
        path = apkbuilds.get(name)
        if path is not None:
            return path

        path = self._search_subpackages(name, apkbuilds)
        return path

    def _search_subpackages(
        self, name: str, apkbuilds: dict
    ) -> Union[Path, None]:
        """Return path to APKBUILD containing named subpackage."""
        for path in apkbuilds.values():
            apkbuild = self._parse_apkbuild(path)
            if name in apkbuild['subpackages']:
                return path
            for prov in apkbuild['provides']:
                if '=' not in prov:
                    # A virtual package, ignore.
                    continue
                prov_name, _ = prov.split('=', 1)
                if name == prov_name:
                    return path
        return None

    def _parse_apkbuild(self, path: Path) -> dict:
        """Return a dict with partial metadata from the APKBUILD."""
        # TODO: run with sudo user, for security purposes.
        cmd = ['bash', '-c', SH_GET_APK_VARS.format(file=str(path))]
        try:
            # This must run with a CWD of where the APKBUILD is located,
            # since it can use relative paths to source other files.
            output = subprocess.run(  # nosec
                cmd, cwd=path.parent, capture_output=True, check=True
            )
            if output.stderr:
                raise subprocess.SubprocessError(output.stderr)
        except subprocess.SubprocessError as e:
            raise exceptions.DownloadError(str(e))
        parsed = output.stdout.decode('utf-8').splitlines()
        # Prepend patch sources with the full path.
        all_sources = parsed[0].split()
        sources = []
        for source in all_sources:
            for scheme in ('http://', 'https://', 'ftp://'):
                if scheme in source:
                    sources.append(source)
                    break
            else:
                abspath = path.resolve().parent / source
                sources.append(f"file://{abspath}")

        # Subpackages can be of the form name:_function. This is where the
        # package is handled in a function in the APKBUILD. We can strip that
        # out for the purposes of matching subpackages.
        subpackages = [item.split(':', 1)[0] for item in parsed[1].split()]
        return {
            'source': sources,
            'subpackages': subpackages,
            'provides': parsed[2].split(),
            'pkgver': parsed[3],
            'pkgrel': parsed[4],
        }

    def _find(self):
        # Main entrypoint from the parent class.
        path = self._find_apkbuild_path(self.name)
        if path is None:
            raise exceptions.SourceNotFound()
        apkbuild = self._parse_apkbuild(path)
        version = f"{apkbuild['pkgver']}-{apkbuild['pkgrel']}"
        # See if there's an 'r' in the release part of the version, and remove
        # it.
        base_version, release = self.version.rsplit('-', 1)
        if release.startswith('r'):
            expect_version = f"{base_version}-{release[1:]}"
        else:
            expect_version = self.version
        if version == expect_version:
            return AlpineDiscoveredSource(apkbuild['source'])
        raise exceptions.SourceNotFound()


class AlpineDiscoveredSource(finder.DiscoveredSource):
    """A discovered Alpine source package."""

    def populate_archive(self, temp_dir, tar):
        # The file name is the last segment of the URL path, unless the source
        # is the special format containing a ::, in which case the part
        # preceding the :: is the file name.
        names = []
        urls = []
        for url in self.urls:
            if '::' in url:
                name, url = url.split('::', 1)
            else:
                name = url.rsplit('/', 1)[-1]
            names.append(name)
            urls.append(url)

        for name, url in zip(names, urls):
            if url.startswith('file:'):
                _, srcfile = url.split('file://', 1)
                arcfile_name = Path(temp_dir) / name
                copyfile(srcfile, arcfile_name)
            elif url.startswith('ftp://'):
                # Requests cannot do FTP, fall back to urllib.
                arcfile_name = self.download_ftp_file(temp_dir, name, url)
            else:
                arcfile_name = self.download_file(
                    temp_dir, name, url, timeout=API_TIMEOUT
                )
            tar.add(arcfile_name, arcname=name, filter=self.reset_tarinfo)

    def download_ftp_file(self, temp_dir, name, url):
        tmp_file_name = pathlib.Path(temp_dir) / name
        with closing(
            # B310 restricts permitted schemes, but we only call with ftp here.
            urllib.request.urlopen(url, timeout=API_TIMEOUT)  # nosec B310
        ) as ftp, open(tmp_file_name, 'wb') as f:
            shutil.copyfileobj(ftp, f)
        return tmp_file_name

    def __repr__(self) -> str:
        return "\n".join(self.urls)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import functools
import pathlib

from launchpadlib.launchpad import Launchpad

from soufi import exceptions, finder

API_TIMEOUT = 30  # seconds


class UbuntuFinder(finder.SourceFinder):
    """Find Ubuntu source files."""

    distro = finder.Distro.ubuntu.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lp_archive = self.get_archive()

    def _find(self):
        build = self.get_build()
        source = self.get_source_from_build(build)
        urls = tuple(sorted(source.sourceFileUrls()))
        return UbuntuDiscoveredSource(urls)

    # NOTE(nic): workaround an apparent bug in Python 3.6 where the default
    #  maxsize for the functools.lru_cache decorator does not work
    @classmethod
    @functools.lru_cache(maxsize=128)
    def get_archive(cls):
        """Retrieve, and cache, the LP distro main archive object."""
        cachedir = pathlib.Path.home().joinpath(".launchpadlib", "cache")
        lp = Launchpad.login_anonymously(
            "soufi",
            "production",
            cachedir,
            version="devel",
            timeout=API_TIMEOUT,
        )
        distribution = lp.distributions[cls.distro]
        return distribution.main_archive

    def get_build(self):
        bins = self.lp_archive.getPublishedBinaries(
            exact_match=True, binary_name=self.name, version=self.version
        )
        try:
            return bins[0].build
        except IndexError:
            raise exceptions.SourceNotFound

    def get_source_from_build(self, build):
        name = build.source_package_name
        ver = build.source_package_version
        sources = self.lp_archive.getPublishedSources(
            exact_match=True, source_name=name, version=ver
        )
        # TODO index error? Can't have a build without a source so this
        # should never fail.
        return sources[0]


class UbuntuDiscoveredSource(finder.DiscoveredSource):
    """A discovered Ubuntu source package."""

    def populate_archive(self, temp_dir, tar):
        # The file name is the last segment of the URL path.
        names = [url.rsplit('/', 1)[-1] for url in self.urls]
        for name, url in zip(names, self.urls):
            arcfile_name = self.download_file(
                temp_dir, name, url, timeout=API_TIMEOUT
            )
            tar.add(arcfile_name, arcname=name, filter=self.reset_tarinfo)

    def __repr__(self):
        return "\n".join(self.urls)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi import exceptions, finder

PUBLIC_PROXY = 'https://proxy.golang.org/'
PROXY_TIMEOUT = 30  # seconds


class GolangFinder(finder.SourceFinder):
    """Find Golang modules.

    A simple HEAD request is made to the Goproxy, which will indicate
    whether the module is available or not.

    The proxy used defaults to the public https://proxy.golang.org/.
    """

    distro = finder.SourceType.go.value

    def __init__(self, *args, goproxy=PUBLIC_PROXY, **kwargs):
        super().__init__(*args, **kwargs)
        self.goproxy = goproxy

    @property
    def original_url(self):
        # It seems as though the proxy only works if the name gets lower-cased.
        # So Github repos such as Shopify/sarama don't work as-is. This Is The
        # Go Way and it must not be questioned.
        return f"{self.goproxy}{self.name.lower()}/@v/{self.version}.zip"

    def _find(self):
        # Main entrypoint from the parent class.
        if self._proxy_find():
            return GolangDiscoveredSource([self.original_url])
        raise exceptions.SourceNotFound()

    def _proxy_find(self):
        response = requests.head(self.original_url, timeout=PROXY_TIMEOUT)
        if response.status_code == requests.codes.not_allowed:
            # HEAD not available; we can try to download it instead and abort
            # before starting the stream.
            response = requests.get(
                self.original_url, stream=True, timeout=PROXY_TIMEOUT
            )
            response.close()
        return response.ok


class GolangDiscoveredSource(finder.DiscoveredSource):
    """A discovered Golang source module."""

    archive_extension = '.zip'
    make_archive = finder.DiscoveredSource.remote_url_is_archive

    # We *might* want to add `Disable-Module-Fetch: true` to the download
    # headers as recommended by the docs. This is left as a future exercise as
    # needed.

    def populate_archive(self, *args, **kwargs):
        pass  # pragma: no cover

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""A simple testing CLI to find and download source."""
import os
import shutil

import click

from soufi import exceptions, finder


class Finder:
    @classmethod
    def find(cls, finder):
        source = finder.find()
        click.echo(source)
        return source

    @classmethod
    def ubuntu(cls, name, version):
        click.echo("Logging in to Launchpad")
        ubuntu_finder = finder.factory(
            "ubuntu", name, version, finder.SourceType.os
        )
        click.echo("Finding source in Launchpad")
        return cls.find(ubuntu_finder)

    @classmethod
    def debian(cls, name, version):
        debian_finder = finder.factory(
            "debian", name, version, finder.SourceType.os
        )
        return cls.find(debian_finder)

    @classmethod
    def npm(cls, name, version):
        npm_finder = finder.factory(
            "npm", name, version, finder.SourceType.npm
        )
        return cls.find(npm_finder)

    @classmethod
    def python(cls, name, version, pyindex=None):
        python_finder = finder.factory(
            "python",
            name=name,
            version=version,
            s_type=finder.SourceType.python,
            pyindex=pyindex,
        )
        return cls.find(python_finder)

    @classmethod
    def centos(cls, name, version, repos=None):
        optimal = 'optimal' in repos
        centos_finder = finder.factory(
            "centos",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            repos=repos,
            optimal_repos=optimal,
        )
        return cls.find(centos_finder)

    @classmethod
    def alpine(cls, name, version, aports_dir):
        alpine_finder = finder.factory(
            "alpine",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            aports_dir=aports_dir,
        )
        return cls.find(alpine_finder)

    @classmethod
    def go(cls, name, version, goproxy):
        go_finder = finder.factory(
            "go",
            name=name,
            version=version,
            s_type=finder.SourceType.go,
            goproxy=goproxy,
        )
        return cls.find(go_finder)

    @classmethod
    def java(cls, name, version):
        java_finder = finder.factory(
            "java",
            name=name,
            version=version,
            s_type=finder.SourceType.java,
        )
        return cls.find(java_finder)


def make_archive_from_discovery_source(disc_src, fname):
    try:
        with disc_src.make_archive() as in_fd, open(fname, 'wb') as out_fd:
            # copyfileobj copies in chunks, so as not to exhaust memory.
            shutil.copyfileobj(in_fd, out_fd)
    except exceptions.DownloadError as e:
        click.echo(str(e))
        click.get_current_context().exit(255)


@click.command()
@click.argument("distro")
@click.argument("name")
@click.argument("version")
@click.option(
    "--pyindex",
    default="https://pypi.org/pypi/",
    help="Python package index if getting Python",
    show_default=True,
)
@click.option(
    "--aports",
    default=None,
    help="Path to a checked-out aports directory if getting Alpine",
    show_default=True,
)
@click.option(
    "--repo",
    default=None,
    multiple=True,
    help="For CentOS, name of repo to use instead of defaults. "
    "Use 'optimal' to use an extended optimal set. May be repeated.",
)
@click.option(
    "--goproxy",
    default="https://proxy.golang.org/",
    help="GOPROXY to use when downloading Golang module source",
    show_default=True,
)
@click.option(
    "--output",
    "-o",
    help="Download the source archive and write to this file name",
    default=None,
)
@click.option(
    "-O",
    "auto_output",
    is_flag=True,
    default=False,
    help="Download the source archive and write to a default file name."
    "The names vary according to the source type and the archive type, "
    "but will generally follow the format: "
    "{name}-{version}.{distro/type}.tar.[gz|xz] "
    "(This option takes precedence over -o/--output)",
)
def main(
    distro, name, version, pyindex, aports, repo, goproxy, output, auto_output
):
    """Find and optionally download source files.

    Given a binary name and version, will find and print the URL(s) to the
    source file(s).

    If the --output option is present, the URLs are all downloaded and
    combined into a LZMA-compressed tar file and written to the file
    name specifed.  If the original source is already an archive then that
    archive is used instead.

    The sources currently supported are 'debian', 'ubuntu', 'centos', 'alpine',
    'java', 'go', 'python' and 'npm', one of which must be specified as the
    DISTRO argument.
    """
    try:
        func = getattr(Finder, distro)
    except AttributeError:
        click.echo(f"{distro} not available")
        click.get_current_context().exit(255)
    if distro == 'alpine' and aports is None:
        click.echo("Must provide --aports for Alpine")
        click.get_current_context().exit(255)
    try:
        if distro == 'python':
            disc_source = func(name, version, pyindex=pyindex)
        elif distro == 'alpine':
            disc_source = func(name, version, aports_dir=aports)
        elif distro == 'centos':
            disc_source = func(name, version, repos=repo)
        elif distro == 'go':
            disc_source = func(name, version, goproxy=goproxy)
        else:
            disc_source = func(name, version)
    except exceptions.SourceNotFound:
        click.echo("source not found")
        click.get_current_context().exit(255)

    if auto_output is not False or output is not None:
        fname = output
        if auto_output:
            name = name.replace(os.sep, '.')
            fname = f"{name}-{version}.{distro}{disc_source.archive_extension}"
        make_archive_from_discovery_source(disc_source, fname)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import abc
import contextlib
import enum
import importlib
import pathlib
import pkgutil
import tarfile
import tempfile
from inspect import isclass
from typing import BinaryIO, ContextManager, Iterable, Union

import requests

from soufi import exceptions


@enum.unique
class SourceType(enum.Enum):
    os = "os"
    python = "python"
    npm = "npm"
    gem = "gem"
    java = "java"
    go = "go"
    nuget = "nuget"


@enum.unique
class Distro(enum.Enum):
    debian = "debian"
    ubuntu = "ubuntu"
    rhel = "rhel"
    centos = "centos"
    alpine = "alpine"


class DiscoveredSource(metaclass=abc.ABCMeta):
    """Base class for all objects implementing a discovered source."""

    # The default 'make_archive' produces .tar.xz files. Override as
    # necessary if make_archive is also overridden.
    archive_extension = '.tar.xz'

    def __init__(self, urls: Iterable[str]):
        self._urls = urls

    @property
    def urls(self):
        return self._urls

    @contextlib.contextmanager
    def make_archive(self) -> ContextManager[BinaryIO]:
        """Make a context manager to a tar archive of all the source URLs.

        Yields a binary file object to the tar archive.

        Any downloaded files are removed after the IO stream is closed by
        exiting this context manager, so it is the caller's responsibility to
        save the stream as necessary.
        """
        # The temp dir is cleaned up once the context manager exits.
        with tempfile.TemporaryDirectory() as target_dir:
            tarfile_fd, tarfile_name = tempfile.mkstemp(dir=target_dir)
            with tarfile.open(name=tarfile_name, mode='w:xz') as tar:
                self.populate_archive(target_dir, tar)
            with open(tarfile_name, 'rb') as fd:
                yield fd

    @contextlib.contextmanager
    def remote_url_is_archive(self) -> ContextManager[BinaryIO]:
        """Replace make_archive when remote URL is already an archive.

        In the case where the URL found for the source is already an archive,
        derived classes can replace the make_archive function with this one
        instead, and it will simply download the remote archive and
        yield its file instead of making a new archive.

        The derived class should do something such as::

            class MyDiscoveredSource(DiscoveredSource):

                make_archive = DiscoveredSource.remote_url_is_archive

                ...

        and define `populate_archive` as a no-op.
        """
        with tempfile.TemporaryDirectory() as target_dir:
            [url] = self.urls
            _, download_file_name = url.rsplit('/', 1)
            tarfile_name = self.download_file(
                target_dir, download_file_name, url
            )
            with open(tarfile_name, 'rb') as fd:
                yield fd

    @abc.abstractmethod
    def populate_archive(self, temp_dir: str, tar: tarfile.TarFile):
        """Populate a TarFile object with downloaded files.

        Derived classes must implement this method.

        :param temp_dir: Name of pre-made temp directory into which the URL
            files can be downloaded.
        :param tar: TarFile object into which the files must be added.
        """
        raise NotImplementedError  # pragma: no cover

    def download_file(
        self,
        target_dir: str,
        target_name: str,
        url: str,
        timeout: int = None,
    ) -> pathlib.Path:
        """Download a file from a URL and place it in a directory.

        Stream-downloads file from <url> and places it as a file named
        <target_name> inside <target_dir>.

        May be called by derived classes to help retrieve files.

        Returns the Path object to the new file.

        Raises a DownloadError upon problems retrieving the remote file.

        NOTE: No decoding is performed on the file, it is saved as raw.
        """
        tmp_file_name = pathlib.Path(target_dir) / target_name
        with requests.get(url, stream=True, timeout=timeout) as response:
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise exceptions.DownloadError(e)
            with open(tmp_file_name, 'wb') as f:
                # Setting chunk_size to None reads whatever size the
                # chunk is as data chunks arrive. This avoids reading
                # the whole file into memory.
                for chunk in response.iter_content(chunk_size=None):
                    f.write(chunk)
        return tmp_file_name

    def reset_tarinfo(self, tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
        """Filter to reset TarInfo fields to remove user details.

        Use as the `filter` parameter to TarFile.add() when populating the
        tar archive.
        """
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = "root"
        return tarinfo


class SourceFinder(metaclass=abc.ABCMeta):
    """Base class for all objects that implement source finding.

    :param name: Name of the source package to find
    :param version: Version of the source package to find
    :param s_type: A SourceType indicating the type of package

    These options may be left unspecified if required and the call to
    `find()` should then specify any that were not passed to `__init__`.
    """

    @property
    @abc.abstractmethod
    def distro(self) -> Union[Distro, str]:
        raise NotImplementedError  # pragma: no cover

    def __init__(
        self, name: str = None, version: str = None, s_type: SourceType = None
    ):
        self.name = name
        self.version = version
        self.s_type = s_type

    def find(
        self, name: str = None, version: str = None, s_type: SourceType = None
    ):
        """Find source and return a DiscoveredSource instance.

        :param name: Name of the source package to find
        :param version: Version of the source package to find
        :param s_type: A SourceType indicating the type of package

        These options may be left unspecified if required and will default
        to the values passed in the constructor. Any value specified here will
        override the value passed in the constructor.
        """
        if name is not None:
            self.name = name
        if version is not None:
            self.version = version
        if s_type is not None:
            self.s_type = s_type

        if None in (self.name, self.version, self.s_type):
            raise ValueError(
                "All of name, version and s_type must have a value"
            )
        return self._find()

    @abc.abstractmethod
    def _find(self) -> DiscoveredSource:
        raise NotImplementedError  # pragma: no cover


class FinderFactory:
    """Factory singleton to return Finder objects.

    Once instantiated, call as:
        factory('<type>', arg1, arg2...)
    Where <type> is a known Finder type (e.g. 'ubuntu') and the rest of the
    args/kwargs are passed directly to the Finder's __init__.
    """

    def __init__(self):
        self._finders = dict()
        # TODO: Make this path configurable.
        import soufi.finders

        for _, module, _ in pkgutil.iter_modules(soufi.finders.__path__):
            mod = importlib.import_module(f"soufi.finders.{module}")
            for _name, obj in mod.__dict__.items():
                if isclass(obj) and issubclass(obj, SourceFinder):
                    # Add class object to our dict.
                    self._finders[obj.distro] = obj

    def __call__(
        self, distro: Union[Distro, str], *args, **kwargs
    ) -> SourceFinder:
        return self._finders[distro](*args, **kwargs)

    @property
    def supported_types(self):
        """Return a list of known Finder types."""
        return list(self._finders.keys())


factory = FinderFactory()



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi import exceptions, finder

SNAPSHOT_API = "https://snapshot.debian.org/"
API_TIMEOUT = 30  # seconds


class DebianFinder(finder.SourceFinder):
    """Find Debian source files.

    Uses the API documented at
    https://salsa.debian.org/snapshot-team/snapshot/blob/master/API
    """

    distro = finder.Distro.debian.value

    def _find(self):
        source_info = self.get_source_info()
        hashes = self.get_hashes(source_info)
        urls = self.get_urls(hashes)
        return DebianDiscoveredSource(urls)

    def get_source_info(self):
        """Return a dict of {name, version} of the source."""
        # This API returns ALL versions of binary packages, you can't
        # put a version in the API. Here, we grab the output and go
        # spelunking.
        url = f"{SNAPSHOT_API}mr/binary/{self.name}/"
        response = requests.get(url, timeout=API_TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = response.json()
        all_versions = data['result']
        for info in all_versions:
            if info['binary_version'] == self.version:
                return dict(name=info['source'], version=info['version'])
        raise exceptions.SourceNotFound

    def get_hashes(self, source_info):
        # Return a list of file hashes as used by Snapshot.
        url = (
            f"{SNAPSHOT_API}mr/package/"
            f"{source_info['name']}/{source_info['version']}/srcfiles"
        )
        response = requests.get(url, timeout=API_TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = response.json()
        try:
            return [r['hash'] for r in data['result']]
        except (IndexError, TypeError):
            raise exceptions.SourceNotFound

    def get_urls(self, hashes):
        # Get file name and return (name, url) pairs
        urls = []
        for hash in hashes:
            url = f"{SNAPSHOT_API}mr/file/{hash}/info"
            response = requests.get(url, timeout=API_TIMEOUT)
            if response.status_code != requests.codes.ok:
                raise exceptions.DownloadError(response.reason)
            data = response.json()
            # The result data is a list. I am unsure what each element
            # of the list can be, but it seems like taking the first
            # returns a valid source file name, which is all we want.
            result = data['result'][0]
            name = result['name']
            urls.append((name, f"{SNAPSHOT_API}file/{hash}"))
        return urls


class DebianDiscoveredSource(finder.DiscoveredSource):
    """A discovered Debian source package.

    This class differs from the Ubuntu one in that it is required
    to separately store the final names of the files we're downloading,
    as they cannot be ascertained from the URL alone.
    """

    def __init__(self, urls):
        names, _urls = zip(*urls)
        super().__init__(_urls)
        self.names = names

    def populate_archive(self, temp_dir, tar):
        for name, url in zip(self.names, self.urls):
            arcfile_name = self.download_file(
                temp_dir, name, url, timeout=API_TIMEOUT
            )
            tar.add(arcfile_name, arcname=name, filter=self.reset_tarinfo)

    def __repr__(self):
        output = []
        for name, url in zip(self.names, self.urls):
            output.append(f"{name}: {url}")
        return "\n".join(output)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi import exceptions, finder

DEFAULT_INDEX = 'https://pypi.org/pypi/'
API_TIMEOUT = 30  # seconds


class PythonFinder(finder.SourceFinder):
    """Find Python source dists.

    Traverses the supplied index, defaulting to the one at pypi.org.

    :param pyindex: optional Python index server; defaults to
        https://pypi.org/pypi/
    """

    distro = finder.SourceType.python.value

    def __init__(self, *args, **kwargs):
        self.index = kwargs.pop('pyindex', DEFAULT_INDEX)
        if self.index[-1] != '/':
            self.index += '/'
        super().__init__(*args, **kwargs)

    def _find(self):
        source_url = self.get_source_url()
        return PythonDiscoveredSource([source_url])

    def get_source_url(self):
        """Get the URL from the JSON info for the Python package."""
        try:
            return self.get_pypi_source_url()
        except (KeyError, exceptions.SourceNotFound):
            if self.index == DEFAULT_INDEX:
                # The default is pypi, no point trying devpi.
                raise
            pass
        return self._get_devpi_source_url()

    def get_pypi_source_url(self):
        """Get URLs for packages that are in a pypi server."""
        url = f"{self.index}{self.name}/{self.version}/json"
        response = requests.get(url, timeout=API_TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = response.json()
        releases = data['releases']
        for version, release_data in releases.items():
            if version != self.version:
                continue
            for item in release_data:
                if item['packagetype'] == 'sdist':
                    return item['url']

        # It should not be possible to get here unless the JSON returned
        # from the index is corrupted.
        raise exceptions.SourceNotFound

    def _get_devpi_source_url(self):
        """Get URLs for packages that are in a devpi server.

        This returns a different structure to pypi format.
        """
        url = f"{self.index}{self.name}/{self.version}"
        headers = {'Accept': 'application/json'}
        response = requests.get(url, headers=headers, timeout=API_TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = response.json()
        result = data['result']
        href = result['+links'][0]['href']
        return href


class PythonDiscoveredSource(finder.DiscoveredSource):
    """A discovered Python sdist package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.tar.gz'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but sdists are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi import exceptions, finder

NPM_REGISTRY = 'https://registry.npmjs.org/'
API_TIMEOUT = 30  # seconds


class NPMFinder(finder.SourceFinder):
    """Find NPM source files.

    Traverses the repository at https://registry.npmjs.org/
    """

    distro = finder.SourceType.npm.value

    def _find(self):
        source_url = self.get_source_url()
        return NPMDiscoveredSource([source_url])

    def get_source_url(self):
        """Get the URL from the JSON info for the NPM package."""
        url = f"{NPM_REGISTRY}{self.name}/{self.version}"
        response = requests.get(url, timeout=API_TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = response.json()
        return data['dist']['tarball']


class NPMDiscoveredSource(finder.DiscoveredSource):
    """A discovered NPM source package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.tar.gz'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but NPMs are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""All exceptions raised by Soufi."""


class SourceNotFound(Exception):
    """Raised when source cannot be located."""

    pass


class DownloadError(Exception):
    """Raised when source cannot be downloaded."""

    pass



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import functools
import re
import urllib
from typing import Iterable

# Bandit reports this as vulnerable but it's OK in lxml now,
# defusedxml's lxml support is deprecated as a result.
import repomd
import requests
from lxml import html  # nosec

from soufi import exceptions, finder

VAULT = "https://vault.centos.org/centos/"
TIMEOUT = 30  # seconds


class CentosFinder(finder.SourceFinder):
    """Find CentOS source files.

    Iterates over the index at https://vault.centos.org/

    :param repos: An iterable of repo names in the [S]Packages directory of
        the source repo. E.g. 'os', 'extras', etc. If not specified,
        a default set of 'os', 'BaseOS', 'updates' and 'extras' are examined.
    :param optimal_repos: (bool) Override repos to select what is considered
        an optimal set to inspect. These are the above defaults, plus:
        'AppStream', 'PowerTools', 'fasttrack'

    The lookup is a 2-stage process:
        1. A "fast" method of examining the index pages and looking for
           files that match the package name and version. This depends on
           the source package name being the same as the binary's and
           will cover ~70% of the look-ups.
        2. A slower fallback method, which downloads the repo's metadata
           and does a direct look-up of the source package name.  This ONLY
           works on version 8 of Centos because the Vault doesn't keep
           binary repodata around for older releases.
    """

    distro = finder.Distro.centos.value

    # List of default source directories in each release. These are the default
    # ones that are inspected for source, and are inspected in the order given
    # here, unless overridden in the constructor.
    # TODO: Allow searching subdirs under these source dirs. The code currently
    #   expects srpms directly under these.
    default_source_dirs = [
        'os/',
        'BaseOS/',
        'updates/',
        'extras/',
    ]

    # Optimal source dirs considered a useful extended set to inspect in
    # addition to the defaults.
    optimal_source_dirs = [
        'AppStream/',
        'os/',
        'BaseOS/',
        'PowerTools',
        'updates/',
        'extras/',
        'fasttrack/',
    ]

    def __init__(
        self,
        *args,
        repos: Iterable[str] = None,
        optimal_repos: bool = False,
        **kwargs,
    ):
        if optimal_repos:
            self.source_dirs = self.optimal_source_dirs
        elif repos is not None:
            self.source_dirs = [f"{repo}/" for repo in repos]
        else:
            self.source_dirs = self.default_source_dirs
        super().__init__(*args, **kwargs)

    def _find(self):
        dirs = self._get_dirs()
        # Start at the latest release and work backwards.
        for dir_ in sorted(dirs, reverse=True):
            for url, dir_listing in self._get_source_dirs_content(dir_):
                ref = self._get_path(dir_listing)
                if ref:
                    target_url = f"{url}/{ref}"
                    return CentosDiscoveredSource([target_url])

        # If we get here, it's likely that the "fast" method has failed
        # because the source package name doesn't match the binary's. We
        # have one more way to look this up using the 'repomd' package,
        # but that only works for v8 packages as the repo data only
        # exists for that release on Vault.
        if 'el8' in self.version:
            url = self._repo_lookup()
            if url:
                return CentosDiscoveredSource([url])
        raise exceptions.SourceNotFound

    def _get_url(self, url):
        response = requests.get(url, timeout=TIMEOUT)
        if response.status_code != requests.codes.ok:
            raise exceptions.DownloadError(response.reason)
        return response.content

    def _get_dirs(self):
        """Get all the possible Vault dirs that could match."""
        url = f"{VAULT}"
        content = self._get_url(url)
        tree = html.fromstring(content)
        dirs = tree.xpath('//td[@class="indexcolname"]/a/text()')
        dirs = [d for d in dirs if d[0].isdigit()]

        # Do very basic checks to see if we can limit searched dirs based on
        # the package version containing a 'elN' hint. Just look for 6/7/8
        # versions for now.
        el = re.search(r'.el([678])', self.version)
        if el is not None:
            dirs = [d for d in dirs if d[0] == el.expand(r'\1')]

        return dirs

    def _get_source_dirs_content(self, release_dir):
        """Yield all content from dirs under release_dir that have source.

        A Release dir is the dir under the top-level VAULT.
        For example, '8/', which when appended to VAULT gives:
            https://vault.centos.org/centos/8/
        Under there, there are many directories that can contain packages,
        and this function yields each in turn.
        """
        url = f"{VAULT}{release_dir}"
        content = self._get_url(url)
        tree = html.fromstring(content)
        dirs = tree.xpath('//td[@class="indexcolname"]/a/text()')
        packages_dir = "SPackages"
        if int(release_dir[0]) < 6 or release_dir in ('6.1/', '6.0/'):
            # Releases up to and including 6.1 use "Packages" as the dir
            # name, and then inexplicably they started using SPackages.
            packages_dir = "Packages"

        # Attempt to only look up in the most common directories, to save time.
        # (sets are not ordered, and we need to keep ordering, so cannot use
        # set intersection here)
        dirs = [d for d in dirs if d in self.source_dirs]

        for dir_ in dirs:
            try:
                _url = f"{url}{dir_}Source/{packages_dir}"
                yield _url, self._get_url(_url)
            except exceptions.DownloadError:
                pass

    def _get_path(self, content):
        """Given a dir on Vault, see if it contains the source HREF."""
        # Grab source directory listing at dir and look for
        # <name>-<version>.src.rpm.
        tree = html.fromstring(content)
        href = f"{self.name}-{self.version}.src.rpm"
        ref = tree.xpath('//a[@href=$href]', href=href)
        if ref:
            return href
        return None

    def _repo_lookup(self):
        for os_dir in self.source_dirs:
            source_url = f"{VAULT}8/{os_dir}Source"
            bin_url = f"{VAULT}8/{os_dir}x86_64/os"
            repo = self._get_repo(bin_url)
            if repo is None:
                break
            for package in repo.findall(self.name):
                if package.evr == self.version:
                    break
            else:
                break
            source_nevra = package.sourcerpm
            # Chop the '.src.rpm' from the end.
            source_evr = source_nevra[:-8]
            source_name, source_version = source_evr.split('-', 1)
            src_repo = self._get_repo(source_url)
            if src_repo is None:
                break
            for spackage in src_repo.findall(source_name):
                if spackage.evr == source_version:
                    return f"{source_url}/{spackage.location}"
        return None

    # Cache repo downloads as they are slow and network-bound.
    @classmethod
    @functools.lru_cache(maxsize=128)
    def _get_repo(cls, url):
        try:
            return repomd.load(url)
        except urllib.error.HTTPError:
            return None


class CentosDiscoveredSource(finder.DiscoveredSource):
    """A discovered Centos source package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.src.rpm'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Src RPMs are already compressed archives, nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



import setuptools

setuptools_version = tuple(map(int, setuptools.__version__.split(".", 2)[:2]))
if setuptools_version < (34, 4):
    raise RuntimeError("setuptools 34.4 or newer is required, detected ",
                       setuptools_version)

if __name__ == "__main__":
    setuptools.setup(
        setup_requires=["pbr>=2.0.0"],
        pbr=True)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests

from soufi import exceptions, finder

MAVEN_SEARCH_URL = 'https://search.maven.org/solrsearch/select'
MAVEN_REPO_URL = 'https://search.maven.org/remotecontent'
API_TIMEOUT = 30  # seconds


class JavaFinder(finder.SourceFinder):
    """Find Java source JARs.

    Uses the SOLR search at https://search.maven.org/
    """

    distro = finder.SourceType.java.value

    def _find(self):
        source_url = self.get_source_url()
        return JavaDiscoveredSource([source_url])

    def get_source_url(self):
        """Construct a URL from the JSON response for the search."""
        params = dict(q=f'a:{self.name} v:{self.version} l:sources', rows=1)
        search = requests.get(
            MAVEN_SEARCH_URL, params=params, timeout=API_TIMEOUT
        )
        # NOTE(nic): empirical observation of throwing "bad" queries at this
        #  thing lead me to believe this will only ever return 200 OK unless
        #  the query string is literally corrupted; if you pass in a
        #  syntactically valid search string for a resource that does not
        #  exist, the response content simply contains a completely
        #  unrelated record.  :-(
        if search.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        data = search.json()

        # Transform the `group` attribute and use it to put together a
        # resource URL from the other data we have.  If it resolves, return
        # it, otherwise the previous search was no good.
        [record] = data['response']['docs']
        group = record['g'].replace('.', '/')
        params = dict(
            filepath=f'{group}/{self.name}/{self.version}/'
            f'{self.name}-{self.version}-sources.jar'
        )
        resource = requests.head(
            MAVEN_REPO_URL,
            params=params,
            allow_redirects=True,
            timeout=API_TIMEOUT,
        )
        if resource.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound
        return resource.url


class JavaDiscoveredSource(finder.DiscoveredSource):
    """A discovered Java source package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.jar'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but JARs are already ZIP files so there is
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]
